/**
 * ==============================================
 *  PMX : Performance Model eXtractor
 * ==============================================
 *
 * (c) Copyright 2014-2015, by Juergen Walter and Contributors.
 *
 * Project Info:   http://descartes.tools/pmx
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.pmx.filter.resourcedemands;

import java.util.HashMap;
import java.util.HashSet;
import java.util.InputMismatchException;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.common.configuration.Configuration;
import kieker.common.record.system.CPUUtilizationRecord;
import kieker.tools.traceAnalysis.filter.AbstractMessageTraceProcessingFilter;

import org.apache.log4j.Logger;

import tools.descartes.librede.LibredeResults;
import tools.descartes.librede.ResultTable;
import tools.descartes.librede.approach.IEstimationApproach;
import tools.descartes.librede.approach.ResponseTimeApproximationApproach;
import tools.descartes.librede.approach.ServiceDemandLawApproach;
import tools.descartes.librede.linalg.LinAlg;
import tools.descartes.librede.linalg.Matrix;
import tools.descartes.librede.linalg.Vector;
import tools.descartes.librede.repository.TimeSeries;
import tools.descartes.pmx.builder.ModelBuilder;
import tools.descartes.pmx.filter.resourcedemands.adapter.LibReDEAdapter;

@Plugin(name = "AbstractResourceDemandFilter", 
description = "Resource demand estimation.", 
outputPorts = {
		@OutputPort(name = AbstractResourceDemandFilter.OUTPUT_PORT_NAME_DEMANDS, 
				description = "Outputs resource demands", 
				eventTypes = {HashMap.class })
})
public abstract class AbstractResourceDemandFilter extends AbstractMessageTraceProcessingFilter {
	
	private static final Scanner SCANNER = new Scanner(System.in);
	public static final String CONFIG_PROPERTY_NAME_OUTPUT_FN = "outputDirectory";
	public static final String OUTPUT_PORT_NAME_DEMANDS = "resourceDemands";
	private static Map<String, TimeSeries> serviceTimeSeriesMap;
	private static Map<String, TimeSeries> resourceTimeSeriesMap;
	private static TimeSeries networkTimeSeries;
	private static Set<String> hosts = new HashSet<String>();
	private final String outputPath;
	private static final Logger log =  Logger.getLogger(AbstractResourceDemandFilter.class);
	
	public static final String INPUT_PORT_NAME_MESSAGE_TRACE = "messageTrace";
	public static final String INPUT_PORT_NAME_EXECUTION_TRACES = "executionTraces";
	public static final String INPUT_PORT_NAME_UTILIZATION = "cpu";

	
	public AbstractResourceDemandFilter(final Configuration configuration,
			final IProjectContext projectContext) {
		super(configuration, projectContext);
		this.outputPath = configuration
				.getPathProperty(CONFIG_PROPERTY_NAME_OUTPUT_FN);
		serviceTimeSeriesMap = new HashMap<String, TimeSeries>();
		resourceTimeSeriesMap = new HashMap<String, TimeSeries>();
	}
	
	@InputPort(name = INPUT_PORT_NAME_UTILIZATION, description = "...", eventTypes = { CPUUtilizationRecord.class })
	public abstract void runUtilizationFilter(final CPUUtilizationRecord record);
//	@InputPort(name = INPUT_PORT_NAME_MESSAGE_TRACE, description = "Extract class-information from monitoring record", eventTypes = { MessageTrace.class })
//	public abstract void runMessageTraceFilter(final MessageTrace trace);


	public static void addNetworkLog(double timestamp, double delay) {
		if(networkTimeSeries== null){
			double[] timeValue = new double[1];
			timeValue[0] = timestamp;
			Vector time = LinAlg.vector(timeValue);

			double[] values = new double[1];
			values[0] = delay;
			Matrix data = LinAlg.matrix(values);
			networkTimeSeries = new TimeSeries(time, data);
		}else{
			networkTimeSeries = networkTimeSeries.addSample(timestamp, delay);
		}
	}
	
	public static synchronized void addResourceLog(double timestamp, String host, String resource, double utilization){
		TimeSeries timeSeries;
		String key = resource+"_"+host;
		if (!resourceTimeSeriesMap.containsKey(key)) {
			double[] timeValue = new double[1];
			timeValue[0] = timestamp;
			Vector time = LinAlg.vector(timeValue);

			double[] values = new double[1];
			values[0] = utilization;
			Matrix data = LinAlg.matrix(values);
			timeSeries = new TimeSeries(time, data);
			resourceTimeSeriesMap.put(key, timeSeries);
		} else {
			timeSeries = resourceTimeSeriesMap.get(key);
			timeSeries = timeSeries.addSample(timestamp, utilization);
			resourceTimeSeriesMap.put(key, timeSeries);
		}
		
	}
	
	public static synchronized void addExecutionLog(double timestamp,
			String interfaceName, String host, double exTime) {
		TimeSeries timeSeries = null;
		String key = interfaceName+ModelBuilder.seperatorChar+host;
		if(exTime < 0 ){
			//log.error("negative execution time. " +host+" "+interfaceName + ": "+exTime);
			//exTime = Math.abs(exTime);
			exTime = 0;
		}
				
		hosts.add(host);
		if (!serviceTimeSeriesMap.containsKey(key)) {
			double[] timeValue = new double[1];
			timeValue[0] = timestamp;
			Vector time = LinAlg.vector(timeValue);

			double[] values = new double[1];
			values[0] = exTime;
			Matrix data = LinAlg.matrix(values);
			timeSeries = new TimeSeries(time, data);
			serviceTimeSeriesMap.put(key, timeSeries);
		} else {
			timeSeries = serviceTimeSeriesMap.get(key);
			timeSeries = timeSeries.addSample(timestamp, exTime);
			serviceTimeSeriesMap.put(key, timeSeries);
		}
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void terminate(final boolean errorBeforeTermination) {
		if(serviceTimeSeriesMap.keySet().isEmpty()){
			log.error("could not extract service times to estimate resource demands");	//No service execution logs could be found
			return;
		}
	
		HashMap<String, Double> resourceDemandMap = new HashMap<String, Double>();
		for (String host : hosts) {
			StringBuffer sb = new StringBuffer();
			for (String service : serviceTimeSeriesMap.keySet()) {
				if (service.endsWith(host)) {
					sb.append(service.replace(ModelBuilder.seperatorChar + host, "") + " | ");
				}
			}
			log.info("Estimate resource demands on [" + host + "]");
			log.info("\tservices: |" + sb.toString());

			/** Run LibReDE */
			LibredeResults estimates = LibReDEAdapter.initAndRunLibrede(host, serviceTimeSeriesMap,
					resourceTimeSeriesMap, outputPath);

			Set<Class<? extends IEstimationApproach>> approaches = estimates.getApproaches();
			Class<? extends IEstimationApproach> approach;
			if (approaches.contains(ServiceDemandLawApproach.class)) {
				approach = ServiceDemandLawApproach.class;
			} else {
				approach = ResponseTimeApproximationApproach.class;
			}

			ResultTable resultTable = estimates.getEstimates(approach, 0);
			Vector x = resultTable.getLastEstimates();
			for (int i = 0; i < x.rows(); i++) {
				// String resourceName = resultTable.getResource(i).getName();
				String serviceName = resultTable.getService(i).getName();
				double rd = x.get(i);
				resourceDemandMap.put(serviceName, rd);
			}
		}

		// Network delay
		if (networkTimeSeries != null) {
			Vector networkDelayVector = networkTimeSeries.getData(0);
			double averageDelay = LinAlg.sum(networkDelayVector).get(0) / networkDelayVector.rows();
			double stdDev = 0;
			for (double d : networkDelayVector.toArray1D()) {
				stdDev += Math.abs(d - averageDelay);
			}
			stdDev = stdDev / networkDelayVector.rows();
			// log.info("\taverageDelay "+averageDelay+", stdDev "+stdDev);
			if (stdDev > 0.5 * averageDelay) {
				log.info("\tstandard deviation for network delays high (" + stdDev + ") compared to average delay ("
						+ averageDelay + "). Maybe extend model with network package size parameters.");
				// log.info("\t==> network model accuracy INsufficient");
			} else {
				log.info("\tstandard deviation for network delays (" + stdDev + ") is ok compared to average delay ("
						+ averageDelay + ")");
				// log.info("\t==> network model accuracy sufficient");
			}
			resourceDemandMap.put("Network", averageDelay);
		}
		super.deliver(AbstractResourceDemandFilter.OUTPUT_PORT_NAME_DEMANDS, resourceDemandMap);
		super.terminate(errorBeforeTermination);
	}
	
	private static int getInputInt() {
			try {
				return SCANNER.nextInt();
			} catch (InputMismatchException e) {
				int result = 2;
				log.info("Input could not be parsed as a number. Value has been set to "+result+"!");
				return result;
			}
	}


}
