/**
 * ==============================================
 *  PMX : Performance Model eXtractor
 * ==============================================
 *
 * (c) Copyright 2014-2015, by Juergen Walter and Contributors.
 *
 * Project Info:   http://descartes.tools/pmx
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.pmx.filter.resourcedemands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.common.configuration.Configuration;
import kieker.common.record.system.CPUUtilizationRecord;
import kieker.tools.traceAnalysis.filter.AbstractMessageTraceProcessingFilter;
import kieker.tools.traceAnalysis.systemModel.AbstractMessage;
import kieker.tools.traceAnalysis.systemModel.AllocationComponent;
import kieker.tools.traceAnalysis.systemModel.Execution;
import kieker.tools.traceAnalysis.systemModel.MessageTrace;

import org.apache.log4j.Logger;

import tools.descartes.librede.units.Time;
import tools.descartes.pmx.builder.ModelBuilder;

@Plugin(name = "InternalResourceDemandFilter", 
description = "Subtracts times for external calls from total time for resource demand estimation.",
outputPorts = { 
		@OutputPort(name = AbstractResourceDemandFilter.OUTPUT_PORT_NAME_DEMANDS,
				description = "Outputs resource demands", eventTypes = { HashMap.class })
})
public class InternalResourceDemandFilter extends AbstractResourceDemandFilter {

	private static final Logger log = Logger
			.getLogger(InternalResourceDemandFilter.class);

	public InternalResourceDemandFilter(final Configuration configuration,
			final IProjectContext projectContext) {
		super(configuration, projectContext);
	}
	
	@InputPort(name = InternalResourceDemandFilter.INPUT_PORT_NAME_UTILIZATION, description = "...", eventTypes = { CPUUtilizationRecord.class })
	public void runUtilizationFilter(final CPUUtilizationRecord record) {
		//TODO Check it is loggingTimestamp or timestamp
		addResourceLog(Time.NANOSECONDS.convertTo(record.getLoggingTimestamp(), Time.SECONDS), record.getHostname(), "CPU", record.getTotalUtilization());
	}

	
//	@InputPort(name = InternalResourceDemandFilter.INPUT_PORT_NAME_MESSAGE_TRACE, description = "Extract class-information from monitoring record", eventTypes = { MessageTrace.class })
//	public void runMessageTraceFilter(final MessageTrace trace) {
//		List<AbstractMessage> messages = trace.getSequenceAsVector();
//		for (AbstractMessage message : messages) {
//			Execution sender = message.getSendingExecution();
//			Execution receiver = message.getReceivingExecution();
//			if (sender.getEss() < receiver.getEss()) {
//				double time = (receiver.getTout() - receiver.getTin()); 
//
//				addExecutionLog(Time.NANOSECONDS.convertTo(receiver.getTin(), Time.SECONDS),
//					receiver.getAllocationComponent().getAssemblyComponent()
//					.getType().getTypeName()
//					+ "." + receiver.getOperation().getSignature().getName(), receiver.getAllocationComponent()
//							.getExecutionContainer().getName(),
//					time);
//			}
//
//		}
//	}


	private static String getKey(Execution x) {
		String key = ""+x.hashCode();
		return key;
	}

	public void terminate(boolean errorBeforeTermination) {
		super.terminate(errorBeforeTermination);
	}

	@Override
	@InputPort(name = AbstractMessageTraceProcessingFilter.INPUT_PORT_NAME_MESSAGE_TRACES, description = "Receives the message traces to be processed",
			eventTypes = { MessageTrace.class })
	public void inputMessageTraces(final MessageTrace mt) {
		Map<Execution, Double> externalCallTime = new HashMap<Execution, Double>();
		Map<Execution, List<Execution>> externalCallMethods = new HashMap<Execution, List<Execution>>();

		List<AbstractMessage> messages = mt.getSequenceAsVector();
		for (AbstractMessage message : messages) {
			Execution sender = message.getSendingExecution();
			Execution receiver = message.getReceivingExecution();
			if (sender.getEss() < receiver.getEss()) {
				if (!externalCallTime.containsKey(sender)) {
					externalCallTime.put(sender, (double) 0);
				}
				if (!externalCallTime.containsKey(receiver)) {
					externalCallTime.put(receiver, (double) 0);
				}
				// Time lost at linking resources
				if (!sender.getAllocationComponent()
						.getExecutionContainer()
						.equals(receiver.getAllocationComponent()
								.getExecutionContainer())) {
					if(!sender.getAllocationComponent().getAssemblyComponent().getName().equals("'Entry'")){
						// sender != receiver 
						//log.info("Network ("+sender.getAllocationComponent().getExecutionContainer().getName()+ " == >" + receiver.getAllocationComponent().getExecutionContainer().getName()+"): "
						//	+ (sender.getTout() - receiver.getTin())
 						//);
					double networkDelay = (receiver.getTin() - sender.getTout());
					double timestamp = sender.getTout();
					addNetworkLog(timestamp, networkDelay);
					}
				}

				double externalTime = (receiver.getTout() - receiver.getTin()); 
				externalCallTime.put(sender, externalCallTime.get(sender) + externalTime);

				List<Execution> list = externalCallMethods.get(sender);
				list = (list == null) ? new ArrayList<Execution>() : list;
				list.add(receiver);
				externalCallMethods.put(sender, list);
			}

		}

		for (Execution execution : externalCallTime.keySet()) {
			if (execution.getAllocationComponent().getAssemblyComponent().getType().getTypeName().contains("Entry")) {
				continue;
			}

			double time = (execution.getTout() - execution.getTin());
			double externalTime = externalCallTime.get(execution);
			if (time < 0) {
				log.error("time < 0: time = " + time);
			}

			//if(true){
			if(externalTime - time > 0.001 * time) {		// measurement >10% uncertain 
				//TODO External calls
				boolean error = true;
				AllocationComponent ac = execution.getAllocationComponent();
				for(Execution sub: externalCallMethods.get(execution)){
					if(!ac.equals(sub.getAllocationComponent())){
						error = false;
					};
				}
				
				if(error){
					log.error("time < external time (trace id "+execution.getTraceId()+") "+execution.getAllocationComponent().getAssemblyComponent().getType().getTypeName() + " "+execution.getOperation().getSignature().getName());
					log.error("\t"+ "time = " + time  );
					log.error("\t"+ "exte = " + externalTime);
					if(externalCallMethods.get(execution) != null){
						for(Execution sub : externalCallMethods.get(execution)){
							log.error("\t"+"       "+(sub.getTout() -sub.getTin()) +" << "+sub.getAllocationComponent().getAssemblyComponent().getType().getTypeName() + " "+ sub.getOperation().getSignature().getName() + "(trace id "+sub.getTraceId()+") ");
						}
					}
				}else{
					log.info(execution.getAllocationComponent().getAssemblyComponent().getType().getTypeName() + " "+execution.getOperation().getSignature().getName() + " has been abortet before external call response (trace id "+execution.getTraceId()+")");
				}
				externalTime = 0.0;
			}

			// Data connection to superclass
			addExecutionLog(Time.NANOSECONDS.convertTo(execution.getTin(), Time.SECONDS),
					execution.getAllocationComponent().getAssemblyComponent()
					.getType().getTypeName()
					+ ModelBuilder.seperatorChar + execution.getOperation().getSignature().getName(), execution.getAllocationComponent()
							.getExecutionContainer().getName(),
					(time - externalTime));
		}
		// //Aufl�sung der Zeimessung < 1ms
		// // Michael Kupperberg aufl�sung von Timern
	}
}
